from abc import ABC, ABCMeta

from django.core.exceptions import ImproperlyConfigured
from django.contrib.auth.models import Group as DjangoGroup

from .permissions import Permission
from .helpers import camel_to_snake


class GroupMetaclass(ABCMeta):
    def __int__(cls):
        return cls.get().pk

    def __str__(cls):
        return cls.__name__

    def __new__(mcls, name, *args, **kwargs):
        cls = super(GroupMetaclass, mcls).__new__(mcls, name, *args, **kwargs)
        if cls.__base__ == ABC:
            return cls

        if not name.endswith('Group'):
            raise ImproperlyConfigured(
                "A Group class's name must end with 'Group'."
            )
        if not hasattr(cls, 'permissions') or\
           not isinstance(cls.permissions, (list, tuple)) or\
           not all(\
                   type(x) == type(Permission) and\
                   issubclass(x, Permission) and\
                   not x is Permission\
                   for x in cls.permissions
           ):
            raise ImproperlyConfigured(
                "A Group class must have a 'permissions' attribute, which must "
                "be a list of Permission classes (except the Permission "
                "base class)."
            )

        if not hasattr(cls, 'group_name'):
            cls.group_name = camel_to_snake(name[:-5])

        return cls


class Group(ABC, metaclass=GroupMetaclass):
    @classmethod
    def get(cls):
        """
        Returns the DB instance representing this group.
        """
        return DjangoGroup.objects.get(
            name=cls.group_name,
        )

    @classmethod
    def _update_or_create(cls):
        """
        !!! This method is private and is intended to only be called from the
        create_permissions management command. Do not call it from your code.

        Creates and returns this group, or updates and returns it based on a
        matching instance.
        """

        return DjangoGroup.objects.update_or_create(
            name=cls.group_name,
        )



import importlib
import inspect
from abc import ABC, ABCMeta

from django.apps import AppConfig
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import Permission as DjangoPermission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db import models

from guardian.backends import ObjectPermissionChecker

from .models import GlobalPermission
from .helpers import camel_to_snake


class PermissionMetaclass(ABCMeta):
    def __iter__(cls):
        # Adding this method enables a Permission subclass to be treated
        # as a list of permissions, which makes checks more idiomatic
        return iter([cls])

    def split(cls, *args, **kwargs):
        # Adding this method is a hacky way of enabling usage of
        # guardian.shortcuts.assign_perm, which expects either a Django
        # Permission object or a string, and for the latter just expects
        # that `.split()` is present on the object. We pretend to be a
        # string in the eyes of `assign_perm` by adding this method.
        return cls.codename.split(*args, **kwargs)

    def __str__(cls):
        return cls.__name__

    def __new__(mcls, name, *args, **kwargs):
        cls = super(PermissionMetaclass, mcls).__new__(mcls, name, *args, **kwargs)
        if cls.__base__ == ABC:
            return cls

        app = cls.__module__.split('.')[0]
        lib = importlib.import_module("{}.apps".format(app))
        app_config = next(
            obj for name, obj in inspect.getmembers(lib)
            if type(obj) == type(AppConfig) and issubclass(obj, AppConfig) \
            and obj != AppConfig
        )
        cls.app_label = app_config.name

        if not hasattr(cls, 'model'):
            raise ImproperlyConfigured(
                "A Permission must contain an explicit `model` attribute. "
                "If you want to define this permission as global, set `model` "
                "to None explicitly."
            )
        if not name.endswith('Permission'):
            raise ImproperlyConfigured(
                "A Permission class's name must end with 'Permission'."
            )
        if not cls.description or not isinstance(cls.description, str):
            raise ImproperlyConfigured(
                "A Permission class must have a 'description' attribute."
            )

        if not hasattr(cls, 'codename'):
            cls.codename = camel_to_snake(name[:-10])

        if cls.model is None:
            cls.codename = '{}.{}'.format(cls.app_label, cls.codename)
            cls.__perm_str__ = 'serious_django_permissions.{}'.format(cls.codename)
        else:
            cls.__perm_str__ = '{}.{}'.format(cls.app_label, cls.codename)

        return cls


class Permission(ABC, metaclass=PermissionMetaclass):
    model = None
    description = ""

    @classmethod
    def _update_or_create(cls):
        """
        !!! This method is private and is intended to only be called from the
        create_permissions management command. Do not call it from your code.

        Creates and returns this permission, or updates and returns it based on a
        matching instance (wrt. codename and content type).
        """
        if cls.model is not None:
            if isinstance(cls.model, str):
                content_type = ContentType.objects.get(
                    app_label=cls.app_label,
                    model=cls.model.lower()
                )
            elif issubclass(cls.model, models.Model):
                content_type = ContentType.objects.get_for_model(
                    cls.model,
                    for_concrete_model=False
                )
            else:
                raise ValueError(
                    "{}.model is not a string or models.Model subclass!".format(cls)
                )

            return DjangoPermission.objects.update_or_create(
                codename=cls.codename,
                content_type=content_type,
                defaults={'name': cls.description},
            )
        else:
            return GlobalPermission.objects.update_or_create(
                codename=cls.codename,
                name=cls.description
            )

    @classmethod
    def get(cls):
        """
        Returns the DB instance representing this permission.
        """
        if cls.model is not None:
            if isinstance(cls.model, str):
                content_type = ContentType.objects.get(
                    app_label=cls.app_label,
                    model=cls.model.lower()
                )
            elif issubclass(cls.model, models.Model):
                content_type = ContentType.objects.get_for_model(
                    cls.model,
                    for_concrete_model=False
                )
            else:
                raise ValueError(
                    "{}.model is not a string or models.Model subclass!".format(cls)
                )

            return DjangoPermission.objects.get(
                codename=cls.codename,
                content_type=content_type,
            )
        else:
            return GlobalPermission.objects.get(
                codename=cls.codename,
                name=cls.description
            )

    @classmethod
    def user_has_perm(cls, user):
        return user.has_perm(cls.__perm_str__)


class PermissionModelBackend(ModelBackend):
    def has_perm(self, user_obj, perm, obj=None):
        if type(perm) == type(Permission) and issubclass(perm, Permission):
            perm_str = perm.__perm_str__
        else:
            perm_str = perm

        if obj is not None:  # use django-guardian check if obj is passed
            check = ObjectPermissionChecker(user_obj)
            return check.has_perm(perm_str, obj)

        return super().has_perm(user_obj, perm_str, obj)



from django.core.management.base import BaseCommand

from ... import helpers


class Command(BaseCommand):
    """
    A command to aggregate the permissions out of all our apps and create them in django models
    """
    help = "create permissions defined anywhere in the current Django project's apps"

    def handle(self, *args, **options):
        """
            This method iterates through all installed apps and searches for
            classes that are inheriting our Permission class.
            If it finds one it creates the permission in the django model
        """
        helpers.create_permissions(self, *args, **options)



from django.db import models
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType


class GlobalPermissionManager(models.Manager):
    def get_queryset(self):
        return super(GlobalPermissionManager, self).get_queryset().filter(
            content_type=ContentType.objects.get_for_model(
                GlobalPermission,
                for_concrete_model=False
            )
        )


class GlobalPermission(Permission):
    """A global permission, not attached to a model"""

    objects = GlobalPermissionManager()

    def save(self, *args, **kwargs):
        self.content_type = ContentType.objects.get_for_model(
            GlobalPermission,
            for_concrete_model=False
        )
        return super().save(*args, **kwargs)

    class Meta:
        proxy = True
